#include <iostream>
#include <iomanip>
#include <fstream>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell-dam.h"
#include "rof.h"

using namespace std;

// ܓxEoxEox30b̒EőlEԒlEŏlE[̃Obhǂ߂B
// 30b(1km)f̊iqTCYłBlatitude(ܓx)̃ObhTCY͖111.2kmœłB
// longitude(ox)̃TCY͈ܓxɂĈقȂ̂ŌvZKvB

void Cell::setElevation (
	int rowStep, // row counter
	int colStep, // column counter
	int nRow, // number of row for the study area
	int nCol, //number of column for the study area
	float SWlat, // South west corner of latitude
	float SWlon, // South west corner of longitude
	int topoScale, // Topography scale. 1 means 3 second.
	float medElevation // elevation at grid center
)
{
	// setting latitude in 'sec' unit
	latitude  = (float)(SWlat*3600 + 3*topoScale*(nRow-rowStep-1) + 1.5*topoScale);	//3: gridcell size in second.

	// setting longitude in 'sec' unit
	longitude = (float)(SWlon*3600 + 3*topoScale*colStep + 1.5*topoScale);

	// setting length (km) for 1 degree in longitude direction
	longitudinalLength = (float)92.6886*topoScale*(float)cos(latitude/3600.0/180.0*3.14);	//92.6886: gridcell size at equator in meter (3s ~ 92.6886 m).

	med = medElevation;
	
	if (rowStep==0 || rowStep==nRow-1 || colStep==0 || colStep==nCol-1) {
		gridKind = -1; // edge of study region
	}
	else{
		gridKind = 1;  // inside of study region
	}
}

/* EאڃObhEObh */
void Cell::setNULL()
{
	down=10;

	downStreamGrid=NULL;
	for(int i=0; i<7; i++) upStreamGrid[i]=NULL;
	ROFCell=NULL;
}


/* o"downStreamGrid"ɒl */
void Cell::setDownStreamGrid(int row, int col, class Cell** cell2)
{
	switch(down){
	case 1:	downStreamGrid = &cell2[row-1][col];
		break;
		
	case 2:	downStreamGrid = &cell2[row-1][col+1];
		break;
		
	case 3:	downStreamGrid = &cell2[row][col+1];
		break;
		
	case 4: downStreamGrid = &cell2[row+1][col+1];
		break;
		
	case 5: downStreamGrid = &cell2[row+1][col];
		break;
		
	case 6: downStreamGrid = &cell2[row+1][col-1];
		break;
		
	case 7: downStreamGrid = &cell2[row][col-1];
		break;
		
	case 8: downStreamGrid = &cell2[row-1][col-1];
		break;
		
	default: break;
	}
}


/* o"upStreamGrid"ɒl */
void Cell::setUpStreamGrid()
{
	for(int i=0; i<7; i++){
		
		if(downStreamGrid->upStreamGrid[i]==NULL){
			downStreamGrid->upStreamGrid[i]=this;
			break;
		}
		
	}
}

/* 㗬Obh̐𐔂 */
void Cell::countNumberOfUpStreamGrid()
{
	numberOfUpStreamGrid = 0;
	
	for(int i=0; i<7; i++){

		if(upStreamGrid[i]!=NULL){
			numberOfUpStreamGrid++;
		}
		else{
			break;
		}
	}
	
}

/* ̃Obh܂܂ROFZ߂ */
void Cell::setROFCell(ROF** fROF, int row_ROF, int col_ROF)
{
	double x=(double)latitude/3600.0;
	double y=(double)longitude/3600.0;
	double distance;
	double distance_min;

	distance_min=100000000000000000000000000000.0;

	for(int i=0; i<row_ROF; i++){
		for(int j=0; j<col_ROF; j++){

			distance = sqrt(pow(x-(double)fROF[i][j].latitude, 2)+pow(y-(double)fROF[i][j].longitude, 2));
			
			if(distance < distance_min){
				
				ROFCell=&fROF[i][j];
				distance_min = distance;
			}
		}
	}

}


// setting parameter for runoff simulation ovZp̃p^̐ݒ
void Cell::Set_parameter(double mMan_slope, double mMan_channel, int thre_basin, 
int x_Div, int t_Div, double theta, int t_step_num, int topoScale)
{
	double latitudeLength;

	latitudeLength = 92.6886*topoScale;

	switch(down){
	case 1:	mL = latitudeLength;
		break;
		
	case 2:	mL = sqrt(pow(latitudeLength, 2) + pow(longitudinalLength, 2));
		break;
		
	case 3:	mL = longitudinalLength;
		break;
		
	case 4: mL = sqrt(pow(latitudeLength, 2) + pow(longitudinalLength, 2));
		break;
		
	case 5: mL = latitudeLength;
		break;
		
	case 6: mL = sqrt(pow(latitudeLength, 2) + pow(longitudinalLength, 2));
		break;
		
	case 7: mL = longitudinalLength;
		break;
		
	case 8: mL = sqrt(pow(latitudeLength, 2) + pow(longitudinalLength, 2));
		break;

	case 0: mL = sqrt(pow(latitudeLength, 2) + pow(longitudinalLength, 2));
		break;
		
	default: break;
	}	
	
	double para_a, para_m; // parameter for setting channel width
	para_a = 1.061722301;
	para_m = 0.685081138;
	
	mB = para_a * pow((basin * latitudeLength/1000.0 * longitudinalLength/1000.0), para_m);
	//mB = 92.6886 * topoScale * longitudinalLength / mL;

	if(downStreamGrid!=NULL){ 
		mG = tan((med - downStreamGrid->med) / mL);

		//if(mG < 0.001) mG = 0.001;
		if(mG < 0.00001) mG = 0.00001;

	}
	else{
		//mG = (float)0.001;
		mG = (float)0.001;		
	}
	mG = atan(mG); mG = sin(mG); // sin̒lɕϊ

	if(basin > thre_basin){
		
		mMan = mMan_channel;
	}
	else{
		
		mMan = mMan_slope;
	}

	mDiv = x_Div;
	mDx = mL / mDiv;
	mDt = t_Div;
	mTheta = theta; 
	mQc = new double[mDiv+1];
	mQn = new double[mDiv+1];
	mAc = new double[mDiv+1];
	mAn = new double[mDiv+1];
	mK = 1.0 / mB;
	mZ = 1.0;
	mM = 1.0 + 0.666667 * mZ;
	mAlpha = 1.0 / mMan * pow(mK, mM-1.0) * sqrt(mG);
	
	mQn_1m = new float*[31];
	for(int i=0; i<31; i++){
		mQn_1m[i] = new float[24];
	}
	for(int i=0; i<31; i++){
		for(int j=0; j<24; j++){
			mQn_1m[i][j] = 0.0;
		}
	}

}


// initial setting for runoff simulation ovZɗpϐEz̏
void Cell::initialize(double ini_lateral, int topoScale)
{
	double up_iniQ;
	double down_iniQ;

	up_iniQ = ini_lateral * 0.001 / 3600 * (basin-1) * 92.6886*topoScale*longitudinalLength;
	down_iniQ = ini_lateral * 0.001 / 3600 * basin * 92.6886*topoScale*longitudinalLength;
	
	
	for(int i=0; i<mDiv+1; i++){

		mQc[i] = (up_iniQ * (mDiv - i) + down_iniQ * i) / mDiv;
		mQn[i] = (up_iniQ * (mDiv - i) + down_iniQ * i) / mDiv;

		mAc[i] = pow(mQc[i]/mAlpha, 1.0/mM);
		mAn[i] = pow(mQn[i]/mAlpha, 1.0/mM);
	}

	mQlc = ini_lateral * 0.001 *  92.6886 * topoScale * longitudinalLength / mL / 3600.0;
	mQln = ini_lateral * 0.001 *  92.6886 * topoScale * longitudinalLength / mL / 3600.0;
}


// getting lateral inflow  ʂ𓾂
void Cell::getLateralRunoff(int n_day, int n_hour, int topoScale)
{
	int c_day, c_hour;
	c_day   = n_day;	
	c_hour  = n_hour - 1;
	
	if(c_hour == 0){

		c_hour = 24;
		c_day--;
	}

	mQlc = ROFCell->value[c_day][c_hour-1]; // previous 1 hour mean data
	mQln = ROFCell->value[n_day][n_hour-1]; // current 1 hour mean data

	// unit change [mm/hr]-->[m2/s]
	mQln = mQln * 0.001 *  92.6886 * topoScale * longitudinalLength / mL / 3600.0;
	mQlc = mQlc * 0.001 *  92.6886 * topoScale * longitudinalLength / mL / 3600.0;
	//
}

// calculate runoff for 1 hour  1ԕ̗ovZ
void Cell::Calculate(int day, int hour, int Ntime)
{
	double dt = (double)mDt;
	float *mQn_1h = new float[Ntime];

	for (int n=0; n < Ntime; n++) {
		// lateral flow intensity.  (m2/s)
		//double wQl = 0.5 * (mQlc + mQln); // mean value last one hour
		//double wQl = mQlc; // mQlc: previous 1 hour mean data
		double wQl = mQln; // mQln: current 1 hour mean data

		// inflow from upstream 㗬[̗ (m3/s)
		double wUpInflow = 0.0;
		for(int i=0; i<numberOfUpStreamGrid; i++){
			// 㗬ɐڑObh̉[̎ԕϗʂZbg
			// setting hourly mean inflow from upper grids
			wUpInflow += (double)upStreamGrid[i]->mQn_1m[day-1][hour-1];
		}

		mQn[0] = wUpInflow; // ^[Qbg̏[痬闬
		mAn[0] = pow(mQn[0]/mAlpha, 1.0/mM); // ^[Qbg̏[ł̐[

		for (int i = 1; i <= mDiv; i++) {
			double wCa = mAlpha*mM*pow(mAc[i-1], mM-1.0);
			// double wCb = mAlpha*mM*pow(mAc[i], mM-1.0);

			double wQ = max(mQc[i], 0.5*(mQc[i-1]+mQn[i-1])+wQl*mDx);
			double wCb = pow(mAlpha, 1.0/mM) * mM * pow(wQ, (mM-1.0)/mM);

			// 		// ȉ̓s pow gȂ̂ňꌩǂɎv邪A
			// 		// mAc[i,i-1]  0 ̂Ƃ floating exception N̂ł
			// 		double wCa = mQc[i-1]/mAc[i-1]*mM[i-1];
			// 		double wCb = mQc[i]/mAc[i]*mM[i];

			// 		double wC = 0.5 * (wCa + wCb);
			//                 if (wC <= EPSQ) wC = EPSQ;

			double wC = 0.5 * (wCa + wCb);

			// ^[Qbg̊evZfʂł̗
			mQn[i] = (mQc[i] + mTheta * wC * (dt/mDx * mQn[i-1] + dt * wQl) 
				- (1. - mTheta) * wC * dt * ((mQc[i] - mQc[i-1])/mDx - wQl)) 
				/ (1. + mTheta * wC * dt / mDx);

			if (mQn[i] < 0.0) {
				mQn[i] = 0.0;
			}

			mAn[i] = pow(mQn[i]/mAlpha, 1.0/mM); // ^[Qbg̊evZfʂł̐[
		}

		// stock simulation at the lower end of slope vZΖʉ[̗ʂL
		mQn_1h[n] = (float)mQn[mDiv];

		// update discharge and water depth ԗʂ̍XV
		for (int i=0; i <= mDiv; i++) {
			mQc[i] = mQn[i];
			mAc[i] = mAn[i];
		}
	}

	float hourlyflow =0.0;
	for (int n=0; n < Ntime; n++) {
		hourlyflow += mQn_1h[n];
	}
	
	mQn_1m[day-1][hour-1] = hourlyflow/Ntime; // stock hourly mean discharge

	delete[] mQn_1h;
}

// return hourly flow calculated by hourly data.
float Cell::GetHourlyFlow(int day, int hour) {
	return (mQn_1m[day][hour]);
}

// return daily flow calculated by hourly data.
float Cell::GetDailyFlow(int day) {
	float dailyflow=0.0;
	for(int hour = 0; hour < 24; hour++){
		dailyflow = dailyflow + mQn_1m[day][hour];
	}
	dailyflow = dailyflow/24.0;
	return (dailyflow);
}

// return hourly maxumum flow in a day.
float Cell::GetHourlyFlowMax(int day) {
	float hourlyflowMax = 0.0;
	for(int hour = 0; hour < 24; hour++) {
		if (mQn_1m[day][hour] >= hourlyflowMax){
			hourlyflowMax = mQn_1m[day][hour];
		}
	}
	return (hourlyflowMax);
}


void Cell::deleteQandA()
{
	delete[] mQc;
	delete[] mQn;
	delete[] mAc;
	delete[] mAn;

	for(int i=0; i < 31; i++) {
		delete[] mQn_1m[i];
	}
}


// set initial condition of dam
void damCell::setDamGrid(
	int wDamRow, 
	int wDamCol, 
	int Row, 
	int Col, 
	class Cell** cell, 
	int wDamSimulationKind
)
{
	damRow = Row - wDamRow; // converting from North starting from 0.
	damCol = wDamCol-1; //  converting from West starting from 0.
	damSimulationKind = wDamSimulationKind; // 0: use observed dam outflow, 1: use dam model
	damGrid = &cell[damRow][damCol];
	damGrid->gridKind = 99;
}


// set initial state variables
void damCell::setDamInitialState(double wStorage)
{
	storage = wStorage; // setting initial storage
}

// set dam model parameters
void damCell::setDamParameter(double wa)
{
	a = wa; // set parameter a
}


// read obserbed 1 month dam outflow
void damCell::readDamOutflow(int iyear, int imonth)
{

	if (damSimulationKind != 0) {
		cout << " Not use dam outflow observed data." << endl;
		return;
	}

	const int dayMonth[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

	int wyear, wmonth, wday;
	float damOutflow;
	char damOutflowFileNamebin[256];

	// Reading daily dam outflow data
	//sprintf(damOutflowFileNamebin, "./input/%04d%02d_damOutflow61-15.dat", iyear, imonth);
	sprintf_s(damOutflowFileNamebin, 256, "./input/%04d%02d_damOutflow61-15.dat", iyear, imonth);
	ifstream fin(damOutflowFileNamebin);
	if (!fin) {
		cout << "Dam outflow file cannot open.\n";
		exit(1);
	}

	// set row and col of dam grid. row begins from south; col begins from west.
	// starting number is 1.

	for (int iday=0; iday < dayMonth[imonth-1]; iday++) {  // day loop
		fin >> wyear >> wmonth >> wday >> damOutflow; // read daily dam outflow data
		for (int ihour=0; ihour < 24; ihour++) { // hour loop
			//cout << " outflow = " << damOutflow << endl;
			damGrid->mQn_1m[iday][ihour] = damOutflow;
		}
	}
	cout << " read input dam outflow data: " << damOutflowFileNamebin << endl;

	fin.close();
}


// calculate dam outflow for 1 hour
void damCell::calculateDamOutflow(int imonth, int day, int hour)
{

	if (damSimulationKind == 0) { // use dam outflow observed data.
		return;
	}

	double wUpInflow = 0.0; // total inflow to dam (m3/s)
	double damOutflow; // outflow form dam (m3/s)

	// inflow from upstream 㗬[̗ (m3/s)
	for(int i=0; i < damGrid->numberOfUpStreamGrid; i++){
		// 㗬ɐڑObh̉[̎ԕϗʂZbg
		// setting hourly mean inflow from upper grids
		wUpInflow += (double)(damGrid->upStreamGrid[i]->mQn_1m[day-1][hour-1]);
	}

	// add raleral inflow
	// lateral flow intensity.  (m2/s)
	wUpInflow += (damGrid->mQln) * (damGrid->mL); 

	damOutflow = a*wUpInflow;

	damGrid->mQn_1m[day-1][hour-1] = (float)damOutflow;
}

